module Ex05 where

import Data.List
import Test.QuickCheck
import Control.Monad.State (State, get, put, evalState, execState)

{-
I was asked in the lecture to write things separately in do, bind and list comprehension
syntax where applicable. I've done this for all the functions in the Practical6.hs
file. If you're still a bit shaky on how it works, have a look at that file and
have a go at trying to convert some of the functions given here into their alternate
notations. Of course, if you have any trouble just ask on the forum and someone will
get back to you (though if it's me specifically you want, maybe try emailing me directly
or tagging me in the post) :) 
-}

-- Task 1. Eight Queens on a Chessboard

-- The queen ♛ is the most powerful piece in the game of chess.
-- Chess is played on an 8x8 checkerboard: each square of the board
-- is labeled by a horizontal and a vertical coordinate ranging from
-- 0 to 7.
--
-- When a queen is placed in one of the squares, it is able to
-- threaten any number of squares vertically, horizontally or
-- diagonally (corner-by-corner) along a straight line.
--
-- For example, the queen Q placed at (1,1) below is able to threaten
-- all the `#` squares, but not the `-` squares.
--
--  0 1 2 3 4 5 6 7
--0 # # # - - - - -
--1 # Q # # # # # #
--2 # # # - - - - -
--3 - # - # - - - -
--4 - # - - # - - -
--5 - # - - - # - -
--6 - # - - - - # -
--7 - # - - - - - #
--
-- The eight queens puzzle is the problem of placing eight chess
-- queens on the checkerboard in such a way that no two of the
-- queens threaten each other's location.
--
-- In a valid solution of the eight queens problem, no two
-- queens can share the same row, column, or diagonal. Since there
-- are eight columns, each column must therefore contain exactly
-- one queen.
--
-- Thus, we will represent our solution as a list `[Int]`
-- of positions: the first entry of the list tells us the location
-- (a row between 0 and 7) of the queen in column 0,
-- the second entry the location of the queen in column 1, and so on.
--
-- A list of length less than 8 denotes a partial solution, where
-- queens were only placed in the first few columns.

type Row = Int
type Queens = [Row] -- position of queen in each column

-- 1.a. Implement a function `extend` that takes a single partial solution
--      containing `n` queens in the first `n` columns, and
--      returns a list of all valid extensions of the partial
--      solution by adding a new queen to the `n+1`st column.
--      If there are no such extensions, the function should return [].
--      HINT 1: write a function that determines safe and threatened
--              row positions in the column.
--      HINT 2: you may want to use the list monad.

type Square = (Int, Int)
safeFrom :: Square -> Square -> Bool
safeFrom (x1, y1) (x2, y2) = dx /= 0 && dy /= 0 && dx /= dy where
    dx = abs $ x1 - x2
    dy = abs $ y1 - y2

safeInNextColumn :: Queens -> Int -> Bool
safeInNextColumn qs c =
    all (\q -> safeFrom q (length qs, c)) (zip [0..] qs) -- q of form (x, y)

extend :: Queens -> [Queens]
extend qs = do
    q <- filter (safeInNextColumn qs) [0..7]
    return $ qs ++ [q]

extend' qs = [qs ++ [q] | q <- filter (safeInNextColumn qs) [0..7]]

-- 1.b. Implement a function that, given a row position `p`,
--      returns the list of all possible solutions to the eight
--      queens problem in which the queen in the first column
--      is placed in the row position `p`.
solutionStartWith :: Int -> [Queens]
solutionStartWith x = sswHelper [x]

sswHelper :: Queens -> [Queens]
sswHelper ps = do
    qs <- extend ps
    case length qs of
        8 -> return qs
        _ -> sswHelper qs

{-
Another way of solving this, which I omitted from the lecture,
is to use a function like the one below which fully extends everything
in a [Queens]. I chose to omit it because people were having a bit
of trouble with things written in the style of safeInNextColumn. This is
also a good way to solve the problem though
-}
fullyExtend :: [Queens] -> [Queens]
fullyExtend [] = fullyExtend [[]]
fullyExtend qss
  | any (\qs -> length qs == 8) qss = qss
  | otherwise = fullyExtend (qss >>= extend)

solutionStartWith' :: Int -> [Queens]
solutionStartWith' x = fullyExtend [[x]]

{-
This prop is a bit slow to run with quickCheck, but it gets
there in the end and shows that these two implementations of
solutionStatWith give the same results
-}
ssw_prop :: Int -> Bool
ssw_prop n = solutionStartWith n == solutionStartWith' n 

-- Task 2. RPN calculator
-- Reverse Polish notation (RPN) is a way of writing arithmetic
-- expressions without parentheses. Apart from the lack
-- of parentheses, its main advantage is that it's easy
-- to type into a calculator.
--
-- Handheld RPN calculators were very popular throughout the 80s,
-- and some people still swear by them.
--
-- Generally, evaluation on an RPN calculator proceeds by maintaining
-- a stack of numbers, and handling each user action one at a time.
-- 1. If the user entered a number (e.g. action `Enter 5`),
--    we push the entered number onto a stack.
-- 2. If the user pressed the button for an arithmetic operation,
--    say `+`, we remove the two topmost number `x` and `y` from
--    the stack, and put the result, (in this case `y + x`) on
--    the stack.
-- 3. There are special buttons: `Clear`, which, when pressed, removes
--    the topmost number from the stack; `AllClear`, which removes
--    everything from the stack, and `Swap`, which swaps the two
--    topmost elements on the stack.

-- Old calculators could store only 4-5 elements on their stack.
-- Here, we allow our stack to be an unbounded list of numbers.
-- The bottom of the stack is always padded with zeroes:
-- so the stack `[]` and the stack `[0,0]` behave identically,
-- as the stacks `[5,4]` and `[5,4,0]` behave identically as well.

-- As an example, the following sequence of actions:
-- `[Enter 3, Enter 7, Enter 5, Arith (-), Arith (*)]`
-- will result in the following final stack state: [6].
-- First, 3, 7 and 5 are put on the stack. Then we subtract
-- 5 from 7, leaving 3 and 2 on the stack. Finally, we multiply
-- these two, which yields 6.

type Stack = [Double]

data UserAction
  = Arith (Double -> Double -> Double)
  | Clear
  | AllClear
  | Swap
  | Enter Double

push :: Double -> State Stack ()
push x =
  get        >>= \xs ->
  put (x:xs)

-- 2.a. Implement the pop operation which removes and returns
--      the top element of the stack. Keep in mind that empty
--      stacks are treated as if they were padded with zeroes.
pop :: State Stack Double
pop = 
    get >>= \ys ->
    case ys of
    [] -> return 0
    (x:xs) -> put xs >> return x
-- The Clear user action lets the user remove the top element
-- of the current stack.
clear :: State Stack ()
clear =
  pop        >>= \_ ->
  return ()

-- 2.b. Implement the AllClear user action, which removes all
--      elements from the stack.
allClear :: State Stack ()
allClear = put []

-- 2.c. Implement the Swap user action, which switches the order
--      of the two topmost elements in the stack. Keep in mind
--      that empty stacks are zero-padded.
swap :: State Stack ()
swap = 
    pop >>= \x ->
    pop >>= \y ->
    push x >>
    push y

{-
This is an example of what not to do when implementing swap.
Whilst this does work, using these lower level state combinators
like get and put means you have to keep track of the data invariants
yourself. This makes the code more error prone and difficult to read
-}
badSwap :: State Stack ()
badSwap = do
    xs <- get
    case length xs of
        0 -> put (0:0:xs)
        1 -> put (0:xs)
        _ -> let (y1:y2:ys) = xs 
             in put (y2:y1:ys)   

-- This just verifies that badSwap is a working implementation of swap
-- and that our problem with it isn't that it doesn't work.
-- The fact that the proposition below holds should not be taken as an
-- endorsement of this implementation :)
swap_prop :: Stack -> Bool
swap_prop s = execState swap s == execState badSwap s 

-- 2.d. Implement the `Arith` user action, which performs the
--      given arithmetic operation on the top two elements of
--      the stack. E.g. if the stack contains `[2,3]`, then
--      `Arith (-)` should result in `[1]`.
performArith :: (Double -> Double -> Double) -> State Stack ()
performArith f = do
    x <- pop
    y <- pop
    push $ f y x

-- 2.e. Implement the `app1` function, which applies the given
--      user action to the current stack.
app1 :: UserAction -> State Stack ()
app1 (Arith f) = performArith f
app1 Clear = clear
app1 AllClear = allClear
app1 Swap = swap
app1 (Enter n) = push n

-- 2.f. Implement the `app` function, which applies the given
--      list of user actions, in order, starting with the head
--      of the list. Once all actions are perforemd, it returns
--      the final state of the stack.
app :: [UserAction] -> State Stack Stack
app [] = get
app (x:xs) =
    app1 x >>
    app xs
{-
Below are some alternate implementations of app, which use functions which may be helpful for
the state monad exercises from Ed. I didn't go through the solutions of these today because
people were more interested in the list monad, but you should give these exercises a go if you're
struggling with state monad stuff :)
-}

{-
Usually when you map stateful computations over a list you want to keep track
of the changes to the state and take them into account in each computation. This
is what the mapM function does, and it has essentially the same logic as above except
functions using mapM can use their return values. The above implementation is closer to
how the mapM_ function works, which is exactly like mapM but it ignores return values
(i.e. it uses >> like above instead of >>=)
-}
app' :: [UserAction] -> State Stack Stack
app' xs = mapM app1 xs >> get

{-
You can also just use the regular map function and then worry about passing state
between things later, which is what the sequence function does. It just maps app1 over
everything and then sequences them all together by chaining together their various stateful
operations. 
-}
app'' :: [UserAction] -> State Stack Stack
app'' xs = sequence (map app1 xs) >> get


-- you can use `runCalc` to test your implementation. E.g.
-- *> runCalc [Enter 3, Enter 7, Enter 5, Arith (-), Arith (*)]
-- [6.0]

-- Try replacing app with app' or app'' to convince yourself that they're
-- the same!
runCalc :: [UserAction] -> Stack
runCalc xs = evalState (app xs) []

